from typing import Dict

from aiothornode.types import ThorChainInfo
from .fixtures import *


def assert_chains_ok(chain_dic: Dict[str, ThorChainInfo], multi):
    assert len(chain_dic) >= 1
    address_set = set()
    for c in chain_dic.values():
        assert len(c.chain) >= 3
        assert c.chain.upper() == c.chain
        assert len(c.pub_key) > 30
        assert c.pub_key.startswith('tthor') or c.pub_key.startswith('thor')
        if multi:
            assert c.gas_rate > 0
        else:
            assert c.gas_rate >= 0
        if c.chain == 'ETH':
            assert len(c.router) > 20

        assert len(c.address) > 10

        address_set.add(c.address)

        assert not c.halted  # I hope you test it when it is not halted, otherwise why are you even doing it?

    assert len(address_set) == len(chain_dic)


@pytest.mark.asyncio
async def test_mccn_chains(chaosnet_connector: ThorConnector):
    chain_dic = await chaosnet_connector.query_chain_info()
    assert_chains_ok(chain_dic, multi=True)


@pytest.mark.asyncio
async def test_sccn_chains(sc_chaosnet_connector: ThorConnector):
    chain_dic = await sc_chaosnet_connector.query_chain_info()
    assert_chains_ok(chain_dic, multi=False)



import asyncio
import logging
import time
import urllib.parse
from operator import itemgetter
from random import Random
from typing import Dict

from aiohttp import ClientSession

from .consensus import consensus_response, first_not_null
from .nodeclient import ThorNodeClient, ThorNodePublicClient
from .types import *


class ThorConnector:
    def __init__(self, env: ThorEnvironment, session: ClientSession, logger=None, rng: Random = None,
                 auto_ban=False):
        self.session = session
        self.env = env
        self.logger = logger or logging.getLogger('ThorConnector')
        self._clients = []
        self._rng = rng or Random()
        self.client_update_period = 120  # sec
        self._last_client_update = 0.0
        self.use_all_nodes_when_updating = True
        self.auto_ban = auto_ban
        self.ban_list = set()

    @property
    def active_ip_addresses(self):
        return [c.node_ip for c in self._clients]

    async def _public_request(self, path: str, is_rpc=False):
        client = ThorNodePublicClient(self.session, logger=self.logger, env=self.env)
        return await client.request(path, is_rpc)

    async def _request(self, path: str, clients: List[ThorNodeClient],
                       consensus=True, post_processor=None,
                       is_rpc=False):
        self.logger.debug(f'Start request to Thor node "{path}"')

        return await self._public_request(path, is_rpc)

    async def _request_consensus(self, path: str, clients: List[ThorNodeClient],
                                 consensus=True, post_processor=None,
                                 is_rpc=False):
        clients = clients or (await self.get_random_clients())

        m, n = self.env.consensus_min, self.env.consensus_total
        if len(clients) < n:
            self.logger.warning(f'Too few nodes in the network!')
            m = 0

        json_responses = await asyncio.gather(*[client.request(path, is_rpc=is_rpc) for client in clients])

        bad_clients = [client.node_ip for client, json_response in zip(clients, json_responses) if
                       json_response is None]
        if self.auto_ban and bad_clients:
            self.logger.info(f'Banning clients: {bad_clients}.')
            self.ban_list.update(set(bad_clients))

        if consensus:
            best_response, ratio = consensus_response(json_responses, consensus_n=m, total_n=n,
                                                      post_processor=post_processor)
        else:
            best_response, ratio = first_not_null(json_responses), 1.0

        if best_response is None:
            ips = [client.node_ip for client in clients]
            self.logger.error(f'No consensus reached between nodes: {ips} for request "{path}"!')
            return None
        else:
            self.logger.debug(f'Success for the request "{path}" consensus: {(ratio * 100.0):.0f}%')
            if isinstance(best_response, dict) and 'code' in best_response:
                raise ThorException(best_response)
            else:
                return best_response

    async def _load_seed_list(self):
        assert self.session
        self.logger.debug(f'Using seed URL: {self.env.seed_url}')
        async with self.session.get(self.env.seed_url) as resp:
            seed_nodes = await resp.json()
            return seed_nodes

    def _new_client(self, ip_addr):
        return ThorNodeClient(ip_addr, self.session, self.env.timeout, self.logger)

    async def _load_active_nodes(self, seed_ips: List[str]):
        self._rng.shuffle(seed_ips)

        for ip in seed_ips:
            client = self._new_client(ip)
            accounts = await client.request(self.env.path_nodes)
            if not accounts:
                continue
            nodes = [ThorNodeAccount.from_json(j) for j in accounts]
            ip_addresses = [node.ip_address for node in nodes if node.is_good]
            self._clients = [self._new_client(ip_address) for ip_address in ip_addresses]
            self.logger.info(f'Active clients: {len(self._clients)} from {len(accounts)}')
            self._last_client_update = time.monotonic()
            return

        self.logger.error('No active clients!')

    @property
    def seconds_from_last_node_update(self):
        return time.monotonic() - self._last_client_update

    async def update_nodes(self):
        pass  # quick fix
        # seeds = await self._load_seed_list()
        # await self._load_active_nodes(seeds)

    def client_list_except_banned(self):
        return [c for c in self._clients if c.node_ip not in self.ban_list]

    async def get_random_clients(self, n=0):
        if not self._last_client_update or time.monotonic() - self._last_client_update > self.client_update_period:
            await self.update_nodes()

        if self.env.consensus_total > len(self._clients):
            return self._clients
        else:
            clients = self.client_list_except_banned() if self.auto_ban else self._clients
            if len(clients) < self.env.consensus_total:
                self.logger.info('Opps! Banned to many clients! Unban them all!')
                self.ban_list = set()
                clients = self._clients

            population_n = self.env.consensus_total if n == 0 else n
            return self._rng.sample(clients, population_n)

    # --- METHODS ----

    async def query_custom_path(self, path, clients=None):
        data = await self._request(path, clients)
        return data

    async def query_node_accounts(self, clients=None, consensus=True) -> List[ThorNodeAccount]:
        data = await self._request(self.env.path_nodes, clients, consensus=consensus,
                                   post_processor=self.post_processor_for_pools)
        return [ThorNodeAccount.from_json(j) for j in data] if data else []

    async def query_queue(self, clients=None, consensus=True) -> ThorQueue:
        data = await self._request(self.env.path_queue, clients, consensus=consensus)
        return ThorQueue.from_json(data)

    async def query_pools(self, height=None, *, clients=None, consensus=True) -> List[ThorPool]:
        if height:
            path = self.env.path_pools_height.format(height=height)
        else:
            path = self.env.path_pools
        data = await self._request(path, clients, consensus=consensus)
        return [ThorPool.from_json(j) for j in data]

    async def query_pool(self, pool: str, height=None, *, clients=None, consensus=True) -> ThorPool:
        if height:
            path = self.env.path_pool_height.format(pool=pool, height=height)
        else:
            path = self.env.path_pool.format(pool=pool)
        data = await self._request(path, clients, consensus=consensus)
        return ThorPool.from_json(data)

    async def query_last_blocks(self, clients=None, consensus=True) -> List[ThorLastBlock]:
        data = await self._request(self.env.path_last_blocks, clients=clients, consensus=consensus)
        return [ThorLastBlock.from_json(j) for j in data] if isinstance(data, list) else [ThorLastBlock.from_json(data)]

    async def query_constants(self, clients=None, consensus=True) -> ThorConstants:
        data = await self._request(self.env.path_constants, clients=clients, consensus=consensus)
        return ThorConstants.from_json(data)

    async def query_mimir(self, clients=None, consensus=True) -> ThorMimir:
        data = await self._request(self.env.path_mimir, clients=clients, consensus=consensus)
        return ThorMimir.from_json(data)

    async def query_chain_info(self, clients=None, consensus=True) -> Dict[str, ThorChainInfo]:
        data = await self._request(self.env.path_inbound_addresses, clients, consensus=consensus)
        if isinstance(data, list):
            info_list = [ThorChainInfo.from_json(j) for j in data]
        else:
            # noinspection PyUnresolvedReferences
            current = data.get('current', {})  # single-chain
            info_list = [ThorChainInfo.from_json(j) for j in current]
        return {info.chain: info for info in info_list}

    async def query_vault(self, vault_type=ThorVault.TYPE_ASGARD, clients=None, consensus=True) -> List[ThorVault]:
        path = self.env.path_vault_asgard if vault_type == ThorVault.TYPE_ASGARD else self.env.path_vault_yggdrasil
        data = await self._request(path, clients, consensus=consensus)
        return [ThorVault.from_json(v) for v in data]

    async def query_balance(self, address: str, clients=None, consensus=True) -> ThorBalances:
        path = self.env.path_balance.format(address=address)
        data = await self._request(path, clients, consensus=consensus)
        return ThorBalances.from_json(data, address)

    async def query_tendermint_block_raw(self, height, clients=None, consensus=True):
        path = self.env.path_block_by_height.format(height=height)
        data = await self._request(path, clients, consensus=consensus, is_rpc=True)
        return data

    async def query_block(self, height, clients=None, consensus=True) -> ThorBalances:
        data = await self.query_tendermint_block_raw(height, clients, consensus)
        return ThorBlock.from_json(data)

    async def query_native_tx(self, tx_hash: str, clients=None, consensus=True, before_hard_fork=False):
        tx_hash = str(tx_hash)
        if not tx_hash.startswith('0x') and not tx_hash.startswith('0X'):
            tx_hash = f'0x{tx_hash}'

        path_pattern = self.env.path_tx_by_hash_old if before_hard_fork else self.env.path_tx_by_hash
        path = path_pattern.format(hash=tx_hash)
        data = await self._request(path, clients, consensus=consensus, is_rpc=True)
        return ThorNativeTX.from_json(data)

    async def query_native_tx_search(self,
                                     query: str,
                                     page: int = 1, per_page: int = 50, order_by='"asc"',
                                     prove=True,
                                     clients=None, consensus=True):
        if not query.startswith('"'):
            query = f'"{query}"'
        if not order_by.startswith('"'):
            order_by = f'"{order_by}"'
        query = urllib.parse.quote_plus(query)
        order_by = urllib.parse.quote_plus(order_by)
        path = self.env.path_tx_search.format(
            query=query,
            prove='true' if prove else 'false',
            page=page,
            per_page=per_page,
            order_by=order_by
        )
        data = await self._request(path, clients, consensus=consensus, is_rpc=True)
        return ThorNativeTXSearchResults.from_json(data)

    async def query_genesis(self, clients=None, consensus=True):
        data = await self._request(self.env.path_genesis, clients, consensus, is_rpc=True)
        return data['result']['genesis'] if data else None

    # --- POST PROCESSORS ----

    @staticmethod
    def post_processor_for_pools(result):
        if not result:
            return result

        r = list(sorted(result, key=itemgetter('node_address')))
        return r



from collections import Counter
from hashlib import sha256
import ujson


def hash_response(r, post_processor=None):
    r = post_processor(r) if post_processor else r
    r_text = ujson.dumps(r, sort_keys=True)  # sorted keys is important for consensus!
    return sha256(r_text.encode('utf-8')).digest()


def sort_dicts_recursive(input):
    if isinstance(input, dict):
        result = {}
        for k, v in sorted(input.items()):
            if isinstance(v, dict):
                result[k] = sort_dicts_recursive(v)
            else:
                result[k] = v
        return result
    elif isinstance(input, (tuple, list)):
        return [sort_dicts_recursive(item) for item in input]
    else:
        return input


def consensus_response(responses, consensus_n, total_n, post_processor=None):
    if not responses:
        return None, 0.0

    hash_dict = {i: hash_response(r, post_processor) for i, r in enumerate(responses) if r is not None}
    counter = Counter(hash_dict.values())
    if not counter:
        return None, 0.0

    most_hash, most_freq = counter.most_common(1)[0]
    if most_freq >= consensus_n:
        best_index = next(i for i, this_hash in hash_dict.items() if this_hash == most_hash)
        return responses[best_index], (most_freq / total_n if total_n else 0.0)
    else:
        return None, 0.0


def first_not_null(responses):
    for r in responses:
        if r:
            return r



import aiohttp

from aiothornode.connector import *
from aiothornode.env import MCCN


async def main():
    env = MCCN.copy()

    async with aiohttp.ClientSession() as session:
        connector = ThorConnector(env, session)

        asgards = await connector.query_vault(ThorVault.TYPE_ASGARD)
        print(asgards)

        print('-----')

        yggdrasils = await connector.query_vault(ThorVault.TYPE_YGGDRASIL)
        print(yggdrasils)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    asyncio.run(main())



import base64
import datetime
import re
import typing
from dataclasses import field
from hashlib import sha256
from typing import List

import ujson
from dateutil.parser import parse as date_parser

from .env import *

THOR_BASE_MULT = 10 ** 8


class ThorException(Exception):
    def __init__(self, j, *args) -> None:
        super().__init__(*args)
        if j and isinstance(j, dict):
            self.code = j.get('code', '')
            self.message = j.get('message', '')
            self.details = j.get('details', [])
        else:
            self.code = -10000
            self.message = 'Failed to decode the error'
            self.details = [
                repr(j)
            ]


@dataclass
class ThorQueue:
    outbound: int = 0
    swap: int = 0
    internal: int = 0
    scheduled_outbound_value: int = 0

    @classmethod
    def from_json(cls, j):
        return cls(
            outbound=int(j.get('outbound', -1)),
            swap=int(j.get('swap', -1)),
            internal=int(j.get('internal', -1)),
            scheduled_outbound_value=int(j.get('scheduled_outbound_value', -1)),
        )

    @property
    def total(self):
        return int(self.outbound) + int(self.swap) + int(self.internal)


@dataclass
class ThorNodeAccount:
    STATUS_STANDBY = 'standby'
    STATUS_ACTIVE = 'active'
    STATUS_READY = 'ready'
    STATUS_WHITELISTED = 'whitelisted'
    STATUS_UNKNOWN = 'unknown'
    STATUS_DISABLED = 'disabled'

    node_address: str = 'thor?'
    status: str = ''
    pub_key_set: dict = field(default_factory=dict)
    validator_cons_pub_key: str = ''
    bond: str = ''
    active_block_height: int = 0
    bond_address: str = ''
    status_since: str = ''
    signer_membership: list = field(default_factory=list)
    requested_to_leave: bool = False
    forced_to_leave: bool = False
    leave_height: int = 0
    ip_address: str = ''
    version: str = ''
    slash_points: int = 0
    jail: dict = field(default_factory=dict)
    current_award: str = ''
    observe_chains: list = field(default_factory=list)
    preflight_status: dict = field(default_factory=dict)
    bond_providers: dict = field(default_factory=dict)

    """
    {
  'node_address': 'thor104gsqwta048e80j909g6y9kkqdjrw0lff866ew',
  'status': 'Whitelisted',
  'pub_key_set': {},
  'validator_cons_pub_key': '',
  'bond': '2',
  'active_block_height': 0,
  'bond_address': 'thor1v6yh9m26tpytqndr368c42lt4ja0898djh5y7l',
  'status_since': 902755,
  'signer_membership': [],
  'requested_to_leave': False,
  'forced_to_leave': False,
  'leave_height': 0,
  'ip_address': '',
  'version': '0.0.0',
  'slash_points': 0,
  'jail': {
    'node_address': 'thor104gsqwta048e80j909g6y9kkqdjrw0lff866ew'
  },
  'current_award': '0',
  'observe_chains': None,
  'preflight_status': {
    'status': 'Standby',
    'reason': 'node account has invalid registered IP address',
    'code': 1
  },
  'bond_providers': {
    'node_address': 'thor104gsqwta048e80j909g6y9kkqdjrw0lff866ew',
    'node_operator_fee': '0',
    'providers': []
  }
}
    """

    @classmethod
    def from_json(cls: 'ThorNodeAccount', j):
        return cls(
            node_address=str(j.get('node_address', '')),
            status=str(j.get('status', '')),
            pub_key_set=j.get('pub_key_set'),
            validator_cons_pub_key=str(j.get('validator_cons_pub_key', '')),
            bond=int(j.get('bond', 0)),
            active_block_height=int(j.get('active_block_height', 0)),
            bond_address=str(j.get('bond_address', '')),
            status_since=int(j.get('status_since', 0)),
            signer_membership=j.get('signer_membership', []),
            requested_to_leave=bool(j.get('requested_to_leave', False)),
            forced_to_leave=bool(j.get('forced_to_leave', False)),
            leave_height=int(j.get('leave_height', 0)),
            ip_address=str(j.get('ip_address', '')),
            version=str(j.get('version', '')),
            slash_points=int(j.get('slash_points', 0)),
            jail=j.get('jail', {}),
            current_award=int(j.get('current_award', 0)),
            observe_chains=j.get('observe_chains', []),
            preflight_status=j.get('preflight_status', {}),
            bond_providers=j.get('bond_providers', {})
        )

    @property
    def preflight_status_reason_and_code(self):
        status = self.preflight_status.get('status', '').lower()
        reason = self.preflight_status.get('reason', '')
        code = self.preflight_status.get('code', 0)
        return status, reason, code

    @property
    def is_good(self):
        status = self.status.lower()
        return status in (self.STATUS_ACTIVE, self.STATUS_WHITELISTED) \
               and not self.requested_to_leave and not self.forced_to_leave and self.ip_address


@dataclass
class ThorLastBlock:
    chain: str = ''
    last_observed_in: int = 0
    last_signed_out: int = 0
    thorchain: int = 0

    @classmethod
    def from_json(cls, j):
        x = cls()
        x.chain = j.get('chain', '')
        x.thorchain = j.get('thorchain', 0)
        x.last_observed_in = j.get('last_observed_in', 0) if 'last_observed_in' in j else j.get('lastobservedin')
        x.last_signed_out = j.get('last_signed_out', 0) if 'last_signed_out' in j else j.get('lastsignedout')
        return x


@dataclass
class ThorPool:
    STATUS_AVAILABLE = 'Available'
    STATUS_BOOTSTRAP = 'Bootstrap'
    STATUS_ENABLED = 'Enabled'

    balance_asset: str = '0'
    balance_rune: str = '0'
    asset: str = ''
    pool_units: str = '0'
    status: str = ''
    synth_units: str = ''
    decimals: str = '0'
    error: str = ''
    pending_inbound_rune: str = '0'
    pending_inbound_asset: str = '0'

    @property
    def balance_asset_int(self):
        return int(self.balance_asset)

    @property
    def balance_rune_int(self):
        return int(self.balance_rune)

    @property
    def pool_units_int(self):
        return int(self.pool_units)

    @property
    def assets_per_rune(self):
        return self.balance_asset_int / self.balance_rune_int

    @property
    def runes_per_asset(self):
        return self.balance_rune_int / self.balance_asset_int

    @classmethod
    def from_json(cls, j):
        return cls(
            balance_asset=j.get('balance_asset', '0'),
            balance_rune=j.get('balance_rune', '0'),
            asset=j.get('asset', ''),
            pool_units=j.get('LP_units', '0') if 'LP_units' in j else j.get('pool_units', '0'),
            status=j.get('status', cls.STATUS_BOOTSTRAP),
            synth_units=j.get('synth_units', '0'),
            decimals=j.get('decimals', '0'),
            error=j.get('error', ''),
            pending_inbound_rune=j.get('pending_inbound_rune', '0'),
            pending_inbound_asset=j.get('pending_inbound_asset', '0'),
        )


@dataclass
class ThorConstants:
    constants: dict = field(default_factory=dict)
    data_types: dict = field(default_factory=dict)

    DATA_TYPES = ('int_64_values', 'bool_values', 'string_values')

    @classmethod
    def from_json(cls, j):
        holder = cls()
        for dt in cls.DATA_TYPES:
            subset = j.get(dt, {})
            holder.data_types[dt] = {}
            for k, v in subset.items():
                holder.constants[k] = v
                holder.data_types[dt][k] = v

        return holder

    def get(self, name, default=None):
        return self.constants.get(name, default)

    def __getitem__(self, item):
        return self.constants[item]


@dataclass
class ThorMimir:
    constants: dict = field(default_factory=dict)

    @classmethod
    def from_json(cls, j: dict):
        holder = cls()
        for k, v in j.items():
            holder.constants[k] = v
        return holder

    def get(self, name, default=None):
        return self.constants.get(name, default)

    def __getitem__(self, item):
        return self.constants[item]


@dataclass
class ThorChainInfo:
    chain: str = ''
    pub_key: str = ''
    address: str = ''
    router: str = ''  # for smart-contract based chains
    halted: bool = False
    gas_rate: int = 0

    @property
    def is_ok(self):
        return self.chain and self.pub_key and self.address

    @classmethod
    def from_json(cls, j):
        return cls(
            chain=j.get('chain', ''),
            pub_key=j.get('pub_key', ''),
            address=j.get('address', ''),
            router=j.get('router', ''),
            halted=bool(j.get('halted', True)),
            gas_rate=int(j.get('gas_rate', 0)),
        )


@dataclass
class ThorCoin:
    asset: str = ''
    amount: int = 0
    decimals: int = 18

    @classmethod
    def from_json(cls, j):
        return cls(
            asset=j.get('asset'),
            amount=int(j.get('amount', 0)),
            decimals=int(j.get('decimals', 18))
        )

    @property
    def amount_float(self):
        return self.amount / (10 ** self.decimals)

    @classmethod
    def from_json_bank(cls, j):
        return cls(
            amount=int(j.get('amount', 0)),
            asset=j.get('denom', ''),
            decimals=8
        )


@dataclass
class ThorRouter:
    chain: str = ''
    router: str = ''

    @classmethod
    def from_json(cls, j):
        return cls(
            chain=j.get('chain', ''),
            router=j.get('router', '')
        )


@dataclass
class ThorAddress:
    chain: str = ''
    address: str = ''

    @classmethod
    def from_json(cls, j):
        return cls(
            chain=j.get('chain', ''),
            address=j.get('address', '')
        )


@dataclass
class ThorVault:
    block_height: int = 0
    pub_key: str = ''
    coins: List[ThorCoin] = field(default_factory=list)
    type: str = ''
    status: str = ''
    status_since: int = 0
    membership: List[str] = field(default_factory=list)
    chains: List[str] = field(default_factory=list)
    inbound_tx_count: int = 0
    outbound_tx_count: int = 0
    routers: List[ThorRouter] = field(default_factory=list)
    addresses: List[ThorAddress] = field(default_factory=list)

    TYPE_YGGDRASIL = 'YggdrasilVault'
    TYPE_ASGARD = 'AsgardVault'

    STATUS_ACTIVE = "Active"
    STATUS_ACTIVE_VAULT = "ActiveVault"
    STATUS_STANDBY = "Standby"
    STATUS_RETIRING = "RetiringVault"

    @property
    def is_active(self):
        return self.status in (self.STATUS_ACTIVE, self.STATUS_ACTIVE_VAULT)

    @classmethod
    def from_json(cls, j):
        return cls(
            block_height=int(j.get('block_height', 0)),
            pub_key=j.get('pub_key', ''),
            coins=[ThorCoin.from_json(coin) for coin in j.get('coins', [])],
            type=j.get('type', ''),
            status=j.get('status', ''),
            status_since=int(j.get('status_since', 0)),
            membership=j.get('membership', []),
            chains=j.get('chains', []),
            inbound_tx_count=int(j.get('inbound_tx_count', 0)),
            outbound_tx_count=int(j.get('outbound_tx_count', 0)),
            routers=[ThorRouter.from_json(r) for r in j.get('routers', [])],
            addresses=[ThorAddress.from_json(a) for a in j.get('addresses', [])],
        )


@dataclass
class ThorBalances:
    RUNE = 'rune'

    height: int
    assets: List[ThorCoin]
    address: str

    @property
    def runes(self):
        for asset in self.assets:
            if asset.asset == self.RUNE:
                return asset.amount
        return 0

    @property
    def runes_float(self):
        return self.runes / THOR_BASE_MULT

    @classmethod
    def from_json(cls, j, address):
        return cls(
            height=0,
            assets=[
                ThorCoin.from_json_bank(item) for item in j.get('balances')
            ],
            address=address
        )

    def find_by_name(self, name):
        candidates = [coin for coin in self.assets if coin.asset == name]
        return candidates[0] if candidates else None


@dataclass
class ThorBlock:
    height: int
    chain_id: str
    time: datetime.datetime
    hash: str
    txs_hashes: List[str]

    @classmethod
    def decode_tx_hash(cls, tx_b64: str):
        decoded = base64.b64decode(tx_b64.encode('utf-8'))
        return sha256(decoded).hexdigest().upper()

    @classmethod
    def from_json(cls, j):
        result = j.get('result', {})
        block = result['block']
        header = block['header']
        time = date_parser(header['time'])

        txs = [
            '0x' + cls.decode_tx_hash(content) for content in block['data']['txs']
        ]

        return cls(
            height=int(header['height']),
            chain_id=header['chain_id'],
            time=time,
            hash=result['block_id']['hash'],
            txs_hashes=txs
        )


class ThorTxAttribute(typing.NamedTuple):
    key: str
    value: str
    index: bool

    @classmethod
    def from_json(cls, j):
        return cls(
            base64.b64decode(j['key']).decode('utf-8'),
            base64.b64decode(j['value']).decode('utf-8'),
            bool(j['index'])
        )


class ThorTxEvent(typing.NamedTuple):
    type: str
    attributes: List[ThorTxAttribute]

    @classmethod
    def from_json(cls, j):
        return cls(
            type=j['type'],
            attributes=[ThorTxAttribute.from_json(a) for a in j['attributes']]
        )

    def value_of(self, key):
        return next(a.value for a in self.attributes if a.key == key)

    @property
    def sender(self):
        return self.value_of('sender')

    @property
    def recipient(self):
        return self.value_of('recipient')

    @property
    def amount(self):
        amt_str = self.value_of('amount')
        value, asset = re.findall(r'[A-Za-z]+|\d+', amt_str)
        return int(value), asset


@dataclass
class ThorNativeTX:
    hash: str
    height: int
    index: int
    code: int
    data: str
    log: List[dict]
    gas_wanted: int
    gas_used: int
    events: List[ThorTxEvent]

    TYPE_SET_MIMIR = 'set_mimir_attr'
    TYPE_ADD_LIQUIDITY = 'add_liquidity'
    TYPE_WITHDRAW = 'withdraw'

    @property
    def type(self):
        return re.sub(r'[^a-zA-Z0-9_]', '', self.data)

    @property
    def transfers(self):
        return [e for e in self.events if e.type == 'transfer']

    @classmethod
    def from_json(cls, j):
        result = j.get('result', j)
        tx_result = result['tx_result']
        data = base64.b64decode(tx_result['data']).decode('utf-8').strip()
        log = ujson.loads(tx_result['log'])
        events = [ThorTxEvent.from_json(e) for e in tx_result['events']]

        return cls(
            hash=result['hash'],
            height=int(result['height']),
            index=int(result['index']),
            code=int(tx_result['code']),
            data=data,
            log=log,
            gas_wanted=int(tx_result['gas_wanted']),
            gas_used=int(tx_result['gas_used']),
            events=events
        )


class ThorNativeTXSearchResults(typing.NamedTuple):
    total_count: int
    txs: List[ThorNativeTX]

    @classmethod
    def from_json(cls, j):
        result = j.get('result', {})
        return cls(
            total_count=int(result.get('total_count', 0)),
            txs=[ThorNativeTX.from_json(tx) for tx in result.get('txs', [])]
        )



from .fixtures import *


@pytest.mark.asyncio
async def test_mccn_search_txs(chaosnet_connector: ThorConnector):
    results = await chaosnet_connector.query_native_tx_search("tx.height=1522004", page=1, per_page=2)
    assert results.total_count == 4
    assert len(results.txs) == 2


@pytest.mark.asyncio
async def test_mccn_block_txs_1(chaosnet_connector: ThorConnector):
    tx_hash = '0xE8510F9636377D66BEC8E263FBFE0B86C92CD3E801794BFFA553C5A9CA42CF09'
    tx = await chaosnet_connector.query_native_tx(tx_hash, consensus=False)
    assert tx.type == 'send'
    assert tx.hash == 'E8510F9636377D66BEC8E263FBFE0B86C92CD3E801794BFFA553C5A9CA42CF09'
    assert len(tx.transfers) == 2
    assert tx.transfers[0].amount == (2000000, 'rune')
    assert tx.transfers[0].recipient == 'thor1dheycdevq39qlkxs2a6wuuzyn4aqxhve4qxtxt'
    assert tx.transfers[0].sender == 'thor179wpxmm5f7asaqwfwnnf8sn3rductlq3ywmrl0'
    assert tx.transfers[0].type == 'transfer'

    assert tx.transfers[1].amount == (4700000000, 'rune')


@pytest.mark.asyncio
async def test_mccn_block_txs_2(chaosnet_connector: ThorConnector):
    tx_hash = 'CE244B951949C3D54DD76C1DAB2FE83C535A1023AE1B9558CF6282AC2C157395'
    tx = await chaosnet_connector.query_native_tx(tx_hash, consensus=False)
    assert tx.type == 'send'
    assert tx.code == 0
    assert tx.height == 1568824
    assert tx.hash == 'CE244B951949C3D54DD76C1DAB2FE83C535A1023AE1B9558CF6282AC2C157395'



from copy import copy
from dataclasses import dataclass


@dataclass
class ThorEnvironment:
    seed_url: str = ''
    midgard_url: str = ''
    thornode_url: str = ''
    rpc_url: str = ''

    timeout: float = 6.0
    consensus_min: int = 2
    consensus_total: int = 3

    path_queue: str = '/thorchain/queue'
    path_nodes: str = '/thorchain/nodes'
    path_pools: str = "/thorchain/pools"
    path_pools_height: str = "/thorchain/pools?height={height}"
    path_pool: str = "/thorchain/pool/{pool}"
    path_pool_height: str = "/thorchain/pool/{pool}?height={height}"

    path_last_blocks: str = "/thorchain/lastblock"
    path_constants: str = "/thorchain/constants"
    path_mimir: str = "/thorchain/mimir"
    path_inbound_addresses: str = "/thorchain/inbound_addresses"
    path_vault_yggdrasil: str = "/thorchain/vaults/yggdrasil"
    path_vault_asgard: str = "/thorchain/vaults/asgard"
    path_balance: str = '/cosmos/bank/v1beta1/balances/{address}'
    path_block_by_height: str = '/block?height={height}'
    path_tx_by_hash: str = '/cosmos/tx/v1beta1/txs/{hash}'
    path_tx_by_hash_old: str = '/tx?hash={hash}'
    path_tx_search: str = '/tx_search?query={query}&prove={prove}&page={page}&per_page={per_page}&order_by={order_by}'

    path_genesis: str = '/genesis'

    kind: str = ''

    def copy(self):
        return copy(self)

    def set_consensus(self, minimum, total):
        assert total >= 1
        assert 1 <= minimum <= total
        self.consensus_total = total
        self.consensus_min = minimum
        return self

    def set_timeout(self, timeout):
        assert timeout > 0.0
        self.timeout = timeout
        return self


class ThorURL:
    class THORNode:
        PUBLIC = 'https://thornode.thorchain.info'
        NINE_REALMS = 'https://thornode.ninerealms.com'
        THORSWAP = 'https://thornode.thorswap.net'

        STAGENET = 'https://stagenet-thornode.ninerealms.com'
        TESTNET = 'https://testnet.thornode.thorchain.info'

    class RPC:
        PUBLIC = 'https://rpc.thorchain.info'
        NINE_REALMS = 'https://rpc.ninerealms.info'
        THORSWAP = 'https://rpc.thorswap.net'

        STAGENET = 'https://stagenet-rpc.ninerealms.com'
        TESTNET = 'https://rpc.thorchain.info'

    class Midgard:
        PUBLIC = 'https://midgard.thorchain.info'
        NINE_REALMS = 'https://midgard.ninerealms.com'
        THORSWAP = 'https://midgard.thorswap.net'

        STAGENET = 'https://stagenet-midgard.ninerealms.com'
        TESTNET = 'https://testnet.midgard.thorchain.info'

    class Seed:
        MAINNET = 'https://seed.thorchain.info'
        TESTNET = 'https://testnet.seed.thorchain.info'


TEST_NET_ENVIRONMENT_MULTI_1 = ThorEnvironment(
    seed_url=ThorURL.Seed.TESTNET,
    midgard_url=ThorURL.Midgard.TESTNET,
    thornode_url=ThorURL.THORNode.TESTNET,
    rpc_url=ThorURL.RPC.TESTNET,
    kind='testnet',
)

MULTICHAIN_STAGENET_ENVIRONMENT = ThorEnvironment(
    midgard_url=ThorURL.Midgard.STAGENET,
    thornode_url=ThorURL.THORNode.STAGENET,
    rpc_url=ThorURL.RPC.STAGENET,
    kind='stagenet',
)

MULTICHAIN_CHAOSNET_ENVIRONMENT = ThorEnvironment(
    seed_url=ThorURL.Seed.MAINNET,
    midgard_url=ThorURL.Midgard.PUBLIC,
    thornode_url=ThorURL.THORNode.PUBLIC,
    rpc_url=ThorURL.RPC.PUBLIC,
    kind='mainnet',
)

MULTICHAIN_CHAOSNET_9R_ENVIRONMENT = ThorEnvironment(
    seed_url=ThorURL.Seed.MAINNET,
    midgard_url=ThorURL.Midgard.NINE_REALMS,
    thornode_url=ThorURL.THORNode.NINE_REALMS,
    rpc_url=ThorURL.RPC.NINE_REALMS,
    kind='mainnet',
)

MULTICHAIN_CHAOSNET_THORSWAP_ENVIRONMENT = ThorEnvironment(
    seed_url=ThorURL.Seed.MAINNET,
    midgard_url=ThorURL.Midgard.THORSWAP,
    thornode_url=ThorURL.THORNode.THORSWAP,
    rpc_url=ThorURL.RPC.THORSWAP,
    kind='mainnet',
)

MCCN = MULTICHAIN_CHAOSNET_ENVIRONMENT  # alias
MCTN = TEST_NET_ENVIRONMENT_MULTI_1  # alias
MCCN_9R = MULTICHAIN_CHAOSNET_9R_ENVIRONMENT  # alias
MCCN_THORSWAP = MULTICHAIN_CHAOSNET_THORSWAP_ENVIRONMENT  # alias



from .fixtures import *


@pytest.mark.asyncio
async def test_mccn_balance(chaosnet_connector: ThorConnector):
    address = 'thor1q9vhc5zz8f097eyx7la4m35wsn7u3vds6sv9kg'
    balance = await chaosnet_connector.query_balance(address, consensus=False)
    # assert balance.height > 0  # no longer available
    assert balance.assets
    assert balance.runes == 341656201145
    assert balance.runes_float == 3416.56201145
    assert balance.address == address
    assert balance.assets

    synth_btc = balance.find_by_name('btc/btc')
    assert synth_btc.amount_float > 0
    assert synth_btc.asset == 'btc/btc'



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='aiothornode',
    version='0.0.22',
    description='THORChain node connection library for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tirinox/aiothornode',
    author='Maksim Koltsov',
    author_email='delevoper@tirinox.ru',
    license='MIT',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'aiohttp>=3.7.3',
        'async-timeout>=3.0.1',
        'attrs>=20.3.0',
        'chardet>=3.0.4',
        'idna>=3.1',
        'multidict>=5.1.0',
        'typing-extensions>=3.7.4.3',
        'python-dateutil>=2.8.2',
        'ujson>=4.0.2',
        'yarl>=1.6.3',
    ]
)



from aiothornode.consensus import sort_dicts_recursive


def test_sort_dicts_recursive_1():
    mixed_dict_1 = {
        "a": 1,
        "c": 3,
        "b": 2,
        "e": 5,
        "d": 4
    }
    sd1 = sort_dicts_recursive(mixed_dict_1)
    assert sd1 == {
        "a": 1,
        "b": 2,
        "c": 3,
        "d": 4,
        "e": 5,
    }
    assert list(sd1.keys()) == ["a", "b", "c", "d", "e"]
    assert list(sd1.keys()) != ["a", "c", "b", "e", "d"]
    assert list(sd1.values()) == [1, 2, 3, 4, 5]
    assert list(sd1.values()) != [1, 3, 2, 5, 4]


def test_sort_dicts_recursive_2():
    mixed_dict_1 = {
        "a": 1,
        "c": 3,
        "b": {
            "zzz": 300,
            "yyy": 200,
            "xxx": 100,
        },
        "e": 5,
        "d": 4
    }
    sd1 = sort_dicts_recursive(mixed_dict_1)

    assert list(sd1.keys()) == ["a", "b", "c", "d", "e"]
    assert list(sd1.values()) == [1, {
            "zzz": 300,
            "yyy": 200,
            "xxx": 100,
        }, 3, 4, 5]

    assert list(sd1['b'].keys()) == ["xxx", "yyy", "zzz"]
    assert list(sd1['b'].keys()) != ["zzz", "yyy", "xxx"]
